package sample;

import java.io.IOException;
import java.net.UnknownHostException;

import javax.security.auth.login.LoginException;

import com.factset.rt.FIDS;
import com.factset.rt.MessageHandler;
import com.factset.rt.RT_Consumer;
import com.factset.rt.RT_Consumer.Subscription;
import com.factset.rt.RT_Message;
import com.factset.rt.RT_OptGreeksFeature;
import com.factset.rt.RT_Request;

public class ProcessGreeksExample {
	private static RT_Consumer consumer;

	private static final MessageHandler msgHandler = new MessageHandler() {

		private String msgType;
		private String delta;
		private String gamma;
		private String vega;
		private String theta;
		private String rho;

		@Override
		public void onMessage(Subscription sub, RT_Message msg) {
			if (msg.isError()) {
				System.err.println("Error: " + msg.getErrorDescription());
				sub.cancel();
				return;
			}

			if (msg.exists(FIDS.MSG_TYPE))
				msgType = msg.getField(FIDS.MSG_TYPE);
			if (msg.exists(FIDS.DELTA))
				delta = msg.getField(FIDS.DELTA);
			if (msg.exists(FIDS.GAMMA))
				gamma = msg.getField(FIDS.GAMMA);
			if (msg.exists(FIDS.VEGA))
				vega = msg.getField(FIDS.VEGA);
			if (msg.exists(FIDS.THETA))
				theta = msg.getField(FIDS.THETA);
			if (msg.exists(FIDS.RHO))
				rho = msg.getField(FIDS.RHO);

			/*
			 * do whatever you like with the extracted information
			 */

			String output = "Received:" + " MsgType = " + msgType + ", Delta = " + delta + ", Gamma = " + gamma
					+ ", Vega = " + vega + ", Theta = " + theta + ", Rho = " + rho;

			System.out.println(output);
			if (msg.isClosed()) {
				sub.cancel();
				System.out.println("Stream Closed by Server...");
			}
		}
	};

	public static void main(String[] args) {
		/**
		 * Establish connection to factset datafeed server
		 */
		String connection = "user:pass@api.df.factset.com";
		consumer = new RT_Consumer(connection);

		try {
			consumer.connect();
		} catch (UnknownHostException e) {
			System.err.println("Error resolving host from: " + connection);
			return;
		} catch (LoginException e) {
			System.err.println("Incorrect username or password: " + connection);
			return;
		} catch (IOException e) {
			System.err.println("Error connecting to server: " + e.getMessage());
			return;
		}

		/**
		 * Connection established. We subscribe to data now
		 */
		consumer.enable(new RT_OptGreeksFeature()); // before requesting
		RT_Request req = new RT_Request("FDS1", "VOD.GB#CWX3D-IFEU");
		Subscription sub = consumer.request(req, msgHandler);

		while (true) {
			consumer.dispatch(-1);
		}
	}
}
