package sample;

import java.awt.Color;
import java.awt.Container;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Insets;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.io.File;
import java.io.IOException;
import java.net.InetAddress;
import java.net.NetworkInterface;
import java.net.SocketException;
import java.net.URL;
import java.net.UnknownHostException;
import java.util.Enumeration;
import java.util.logging.FileHandler;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.logging.SimpleFormatter;
import java.util.prefs.Preferences;

import javax.security.auth.login.LoginException;
import javax.swing.BorderFactory;
import javax.swing.ImageIcon;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JComboBox;
import javax.swing.JComponent;
import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JScrollPane;
import javax.swing.JTabbedPane;
import javax.swing.SwingUtilities;
import javax.swing.UIManager;
import javax.swing.WindowConstants;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import javax.swing.filechooser.FileNameExtensionFilter;

import org.xml.sax.SAXException;

import sample.SwingQuoteTable.FidCell;

import com.factset.rt.ControlHandler;
import com.factset.rt.MessageHandler;
import com.factset.rt.NotifyHandler;
import com.factset.rt.RT_CalculatorFeature;
import com.factset.rt.RT_Consumer;
import com.factset.rt.RT_Consumer.Subscription;
import com.factset.rt.RT_FidField;
import com.factset.rt.RT_FieldMap;
import com.factset.rt.RT_Message;
import com.factset.rt.RT_OptGreeksFeature;
import com.factset.rt.RT_Request;
import javax.swing.JTextField;
import javax.swing.JTextPane;

public class SwingQuote implements MessageHandler, ControlHandler, NotifyHandler {

    /*
     * Constants
     */
    private final Color ERROR_TEXT_COLOR = new Color(180, 0, 0);

    /*
     * UI components
     */
    private JFrame frame;
    private JTabbedPane uriTabPane;
    private JComboBox enterpriseUriCombo;
    private JComboBox workstationUriCombo;
    private JComboBox serviceCombo;
    private JComboBox keyCombo;
    private JButton connectButton;
    private JButton disconnectButton;
    private JCheckBox otpCheck;
    private JCheckBox otpIndivCheck;
    private JCheckBox greeksCheck;
    private JButton requestButton;
    private JButton cancelButton;
    private JButton closeButton;
    private JLabel statusLabel;
    private JLabel updatesLabel;
    private ImageIcon blankIcon;
    private ImageIcon greenIcon;
    private ImageIcon greyIcon;
    private SwingQuoteTable msgTable;

    /*
     * Application elements
     */
    private RT_Consumer consumer;
    private RT_FieldMap map;
    private Subscription requestedSub = null;
    private final Logger logger = Logger.getLogger("SwingQuote");
    private final Preferences prefs = Preferences.userNodeForPackage(getClass());
    private int numUpdates = 0;

    private final Runnable dispatchRunnable = new Runnable() {
        @Override
        public void run() {
            consumer.dispatch();
        }
    };
    private JComboBox proxyUrlCombo;
    private JComboBox proxyUsernameCombo;
    private JTextField proxyPasswordTxt;
    private JTextPane txtProxy;

    @Override
    public void onNotify() {
        SwingUtilities.invokeLater(dispatchRunnable);
    }

    @Override
    public void onMessage(Subscription sub, RT_Message msg) {
        if (!sub.equals(requestedSub)) {
            /*
             * Not possible, since we only request one data item at a time.
             */
            logger.severe("Subscription Mismatch! Received sub: " + sub + ", Requested sub: " + requestedSub);
            sub.cancel();
            cancelButton.setEnabled(false);
            requestedSub = null;
            numUpdates = 0;
            return;
        }

        /*
         * Set our status text
         */
        if (msg.isError()) {
            msgTable.setRowCount(0);
            updatesLabel.setForeground(ERROR_TEXT_COLOR);
            updatesLabel.setText("Stream Closed: " + msg.getError() + " " + msg.getErrorDescription());

        } else if (msg.isClosed()) {
            updatesLabel.setForeground(ERROR_TEXT_COLOR);
            updatesLabel.setText("Stream Closed by Server: " + msg.getStaleReason());
        } else if (msg.isStale()) {
            updatesLabel.setForeground(ERROR_TEXT_COLOR);
            updatesLabel.setText("Stream Stale: " + msg.getStaleReason());
        } else {
            updatesLabel.setForeground(Color.BLACK);
            updatesLabel.setText("Stream OK, NumUpdates: " + numUpdates);
        }

        if (msg.isResponse()) {
            // Its the first message
            statusLabel.setText(
                    "Key: " + msg.getKey() + "  NumFids: " + msg.count() + "  PermID: " + msg.getPermissions());

            msgTable.setRowCount(msg.count());
            int i = 0;
            for (RT_FidField field : msg) {
                String fieldName;
                if (map != null)
                    fieldName = map.getFieldInfo(field.getFid()).getName();
                else
                    fieldName = "{NOMAP}";
                msgTable.setRow(i, field.getFid(), blankIcon, fieldName, field.getValue());
                ++i;
            }
        } else {
            // Its an update
            for (int i = 0; i < msgTable.getRowCount(); ++i) {
                FidCell fidCell = msgTable.getFidCellAt(i, 0);
                if (msg.exists(fidCell.fid)) {
                    fidCell.icon = greenIcon;
                    msgTable.setValueAt(fidCell, i, 0);
                    msgTable.setValueAt(msg.getField(fidCell.fid), i, 2);
                } else if (fidCell.icon == greenIcon) {
                    fidCell.icon = greyIcon;
                    msgTable.setValueAt(fidCell, i, 0);
                }
            }
        }

        if (msg.isClosed()) {
            /*
             * The server closed the stream (possibly due to an error). Clean up our
             * subscription, and reset our state.
             */
            sub.cancel();
            cancelButton.setEnabled(false);
            requestedSub = null;
            numUpdates = 0;
        } else {
            ++numUpdates;
        }
    }

    @Override
    public void onControl(boolean isConnected, RT_Message controlMsg) {
        if (controlMsg.getKey().equals("CONNECTED")) {
            String[] services = consumer.getServices();
            for (int i=0; i<services.length; ++i)
                serviceCombo.addItem(services[i]);

            requestButton.setEnabled(true);
            statusLabel.setText("Connected to: " +
                    (String)((JComboBox)uriTabPane.getSelectedComponent()).getSelectedItem());
        }
        else if (controlMsg.getKey().equals("DISCONNECTED")) {
            serviceCombo.removeAllItems();
            requestButton.setEnabled(false);
            statusLabel.setText("Lost connection... Retrying");
        }
        else if (controlMsg.getKey().equals("SERVICE_ENABLE")) {
            for (RT_FidField field : controlMsg) {
                serviceCombo.addItem(field.getValue());
            }
        }
        else if (controlMsg.getKey().equals("SERVICE_DISCONNECT")) {
            for (RT_FidField field : controlMsg) {
                serviceCombo.removeItem(field.getValue());
            }
        }
    }

    private void connect() {
        String uri = (String)((JComboBox)uriTabPane.getSelectedComponent()).getSelectedItem();
        if (uri == null) { uri = "";}

        try {
            if (uriTabPane.getSelectedComponent() == enterpriseUriCombo) {
                if(otpCheck.isSelected()) {
                    JOptionPane.showMessageDialog(frame, "Please select a file named <keyId>.data and containing <your_key> on the first line and <your_counter> in decimal format on the second.", "Select Key/Counter File", JOptionPane.INFORMATION_MESSAGE);
                    JFileChooser chooser = new JFileChooser();
                    FileNameExtensionFilter filter = new FileNameExtensionFilter("Data Files (*.data)", "data");
                    chooser.setFileFilter(filter);
                    chooser.setCurrentDirectory(new File(System.getProperty("user.dir")));
                    if(chooser.showOpenDialog(frame)!=JFileChooser.APPROVE_OPTION) {
                        return;
                    }
                    String fullPath = chooser.getSelectedFile().getAbsolutePath();
                    String path = "";
                    String deviceId = "";
                    int sepLoc = fullPath.lastIndexOf("\\");
                    if(sepLoc!=-1) {
                        path = fullPath.substring(0, sepLoc);
                        deviceId = fullPath.substring(sepLoc+1);
                    } else {
                        deviceId = fullPath;
                    }
                    int dotLoc = deviceId.lastIndexOf(".");
                    if(dotLoc!=-1) {
                        deviceId = deviceId.substring(0, dotLoc);
                    }
                    String user = "";
                    String host = "";
                    int atLoc = uri.indexOf("@");
                    if(atLoc!=-1) {
                        user = uri.substring(0, atLoc);
                        host = uri.substring(atLoc+1);
                    }
                    int colLoc = user.indexOf(":");
                    if(colLoc!=-1) {
                        user = user.substring(0, colLoc);
                    }
                    if(host.isEmpty() || user.isEmpty()) {
                        JOptionPane.showMessageDialog(frame, "Host or user cannot be discerned. Ensure connection uri is in the form <user>-<serial>@<host>", "Error", JOptionPane.ERROR_MESSAGE);
                        return;
                    }
                    if(deviceId.isEmpty()) {
                        JOptionPane.showMessageDialog(frame, "Error determining keyId from file. Ensure key/counter file is named <keyId>.data", "Error", JOptionPane.ERROR_MESSAGE);
                        return;
                    }
                    boolean machineAccount = !otpIndivCheck.isSelected();
                    consumer = new RT_Consumer(host, user, deviceId, null, null, path, false, machineAccount);
                } else {
                    consumer = new RT_Consumer(uri);
                }
                // proxy settings
                String proxyUrl = (String)proxyUrlCombo.getSelectedItem();
                if (proxyUrl != null && !proxyUrl.isEmpty()) {
                    String proxyUsername = (String)proxyUsernameCombo.getSelectedItem();
                    String proxyPassword = proxyPasswordTxt.getText();
                    if (proxyUsername != null && !proxyUsername.isEmpty()) {
                        consumer.setHttpProxy(proxyUrl, proxyUsername, proxyPassword);
                    } else {
                        consumer.setHttpProxy(proxyUrl);
                    }
                }
            } else {
                String user;
                String pass = "";
                int delimIdx = uri.indexOf(':');
                if (delimIdx == -1)
                    user = uri;
                else {
                    user = uri.substring(0, delimIdx);
                    pass = uri.substring(delimIdx+1);
                }
                consumer = RT_Consumer.newWorkstationInstance(user, pass);
            }
            
            if(greeksCheck.isSelected()) {
            	consumer.enable(new RT_OptGreeksFeature());
			}
            
            RT_CalculatorFeature feature = new RT_CalculatorFeature(map);
            feature.addNetChangeCalculator();
            feature.addPctChangeCalculator();
            feature.addMidNetChangeCalculator();
            feature.addMidPctChangeCalculator();
            consumer.enable(feature);
            
            consumer.setNotifyHandler(this);
            consumer.setControlHandler(this);           
            logger.info("Connecting to " + uri);
            consumer.connect();
        } catch (UnknownHostException e) {
            String errMsg;
            if (uriTabPane.getSelectedComponent() == enterpriseUriCombo)
                errMsg = "Error resolving host from the connection string.";
            else
                errMsg = "No running workstation detected.";

            JOptionPane.showMessageDialog(frame, errMsg, "Error",
                    JOptionPane.ERROR_MESSAGE);
            return;
        } catch (IOException e) {
            logger.severe("Error connecting to DataFeed server: " +
                    e.getMessage());
            JOptionPane.showMessageDialog(frame, "Error connecting:\n" +
                    e.getMessage(), "Error", JOptionPane.ERROR_MESSAGE);
            return;
        } catch (LoginException e) {
            JOptionPane.showMessageDialog(frame,
                    "Incorrect username or password.", "Error",
                    JOptionPane.ERROR_MESSAGE);
            return;
        }

        addItemToCombo((JComboBox)uriTabPane.getSelectedComponent(), uri);
        uriTabPane.getSelectedComponent().setEnabled(false);
        uriTabPane.setEnabledAt((uriTabPane.getSelectedIndex() == 0 ? 1 : 0), false);

        for (String service : consumer.getServices())
            serviceCombo.addItem(service);
        serviceCombo.setSelectedItem(prefs.get("SERVICE", ""));

        connectButton.setEnabled(false);
        disconnectButton.setEnabled(true);
        requestButton.setEnabled(true);
        keyCombo.setEnabled(true);
        greeksCheck.setEnabled(false);
        otpCheck.setEnabled(false);
        setProxyItems(false);
        frame.getRootPane().setDefaultButton(requestButton);

        statusLabel.setText("Connected to: " + uri);
    }

    private void disconnect() {
        consumer.disconnect();

        resetGUIState();
        requestedSub = null;
        numUpdates = 0;
    }

    private void request() {
        cancel();
        String service = (String)serviceCombo.getSelectedItem();
        String key = (String)keyCombo.getSelectedItem();
        if (key == null)
            key = "";

        msgTable.setRowCount(0);

        addItemToCombo(keyCombo, key);
        prefs.put("SERVICE", service);

        RT_Request req;
        if(service.equals("FDS_PERM")) {
            req = new RT_Request(service, key, false, "", getUserIPAddresses());
        } else {
            req = new RT_Request(service, key);
        }
        requestedSub = consumer.request(req, this);

        cancelButton.setEnabled(true);
        statusLabel.setText("Request sent for " + service + "::" + key);
    }

    private String getUserIPAddresses() {
        String result = "";
        Enumeration<NetworkInterface> interfaces;
        try {
            interfaces = NetworkInterface.getNetworkInterfaces();
            while(interfaces.hasMoreElements()) {
                NetworkInterface n = interfaces.nextElement();
                Enumeration<InetAddress> iNetAddrs = n.getInetAddresses();
                while(iNetAddrs.hasMoreElements()) {
                    InetAddress i = iNetAddrs.nextElement();
                    if(result.isEmpty()) {
                        result = i.getHostAddress();
                    } else {
                        result += ", " + i.getHostAddress();
                    }
                }
            }
            return result;
        } catch (SocketException e) {
            e.printStackTrace();
            return "";
        }
    }

    private void cancel() {
        if (requestedSub != null) {
            requestedSub.cancel();
            requestedSub = null;

            cancelButton.setEnabled(false);
            numUpdates = 0;

            statusLabel.setText("Cancelled subscription");
            updatesLabel.setForeground(Color.BLACK);
            updatesLabel.setText("NumUpdates: 0");
        }
    }

    private void resetGUIState() {
        serviceCombo.removeAllItems();
        uriTabPane.getSelectedComponent().setEnabled(true);
        uriTabPane.setEnabled(true);
        uriTabPane.setEnabledAt((uriTabPane.getSelectedIndex() == 0 ? 1 : 0), true);

        keyCombo.setEnabled(false);
        connectButton.setEnabled(true);
        disconnectButton.setEnabled(false);
        requestButton.setEnabled(false);
        cancelButton.setEnabled(false);

        greeksCheck.setEnabled(true);
        otpCheck.setEnabled(true);
        setProxyItems(true);
        statusLabel.setText("Disconnected");
        updatesLabel.setForeground(Color.BLACK);
        updatesLabel.setText("NumUpdates: 0");
    }

    private void addComponentsToPane() {
        Container pane = frame.getContentPane();
        ((JComponent)pane).setBorder(BorderFactory.createEmptyBorder(0, 10,
                10, 10));
        GridBagLayout gridBagLayout = new GridBagLayout();
        gridBagLayout.rowWeights = new double[]{0.0, 0.0, 0.0, 0.0, 0.0, 0.0};
        gridBagLayout.columnWeights = new double[]{1.0, 1.0, 0.0, 1.0, 1.0, 0.0};
        pane.setLayout(gridBagLayout);

        enterpriseUriCombo = new JComboBox();
        enterpriseUriCombo.setPreferredSize(new Dimension(225, 25));
        enterpriseUriCombo.setMinimumSize(new Dimension(225, 25));
        enterpriseUriCombo.setEditable(true);
        getItemsFromPrefs(enterpriseUriCombo, "ENT_URI", 3);

        workstationUriCombo = new JComboBox();
        workstationUriCombo.setEditable(true);
        getItemsFromPrefs(workstationUriCombo, "WKS_URI", 3);

        GridBagConstraints uriCon = new GridBagConstraints();
        uriCon.fill = GridBagConstraints.HORIZONTAL;
        uriCon.anchor = GridBagConstraints.LINE_START;
        uriCon.gridx = 0;
        uriCon.gridy = 0;
        uriCon.gridwidth = 3;
        uriCon.weightx = 1;
        uriCon.insets = new Insets(0, 0, 5, 5);
        uriTabPane = new JTabbedPane();
        uriTabPane.setTabLayoutPolicy(JTabbedPane.SCROLL_TAB_LAYOUT);
        uriTabPane.addTab("Enterprise", enterpriseUriCombo);
        uriTabPane.addTab("Workstation", workstationUriCombo);
        uriTabPane.addChangeListener(new ChangeListener() {
            @Override
            public void stateChanged(ChangeEvent e) {
                if(uriTabPane.getSelectedComponent() == enterpriseUriCombo) {
                    otpCheck.setEnabled(true);
                    greeksCheck.setEnabled(true);
                    setProxyItems(true);
                } else {
                    otpCheck.setEnabled(false);
                    otpCheck.setSelected(false);
                    otpIndivCheck.setEnabled(false);
                    otpIndivCheck.setSelected(false);
                    setProxyItems(false);
                }
            }
        });
        pane.add(uriTabPane, uriCon);

        GridBagConstraints serviceCon = new GridBagConstraints();
        serviceCon.insets = new Insets(0, 0, 5, 5);
        serviceCon.fill = GridBagConstraints.HORIZONTAL;
        serviceCon.gridx = 0;
        serviceCon.gridy = 1;
        serviceCombo = new JComboBox();
        serviceCombo.setFont(new Font("Tahoma", Font.PLAIN, 11));
        pane.add(serviceCombo, serviceCon);

        GridBagConstraints keyCon = new GridBagConstraints();
        keyCon.fill = GridBagConstraints.HORIZONTAL;
        keyCon.anchor = GridBagConstraints.LINE_START;
        keyCon.gridx = 1;
        keyCon.gridy = 1;
        keyCon.gridwidth = 2;
        keyCon.insets = new Insets(0, 5, 5, 5);
        keyCombo = new JComboBox();
        keyCombo.setEditable(true);
        keyCombo.setEnabled(false);
        getItemsFromPrefs(keyCombo, "KEY", 3);
        pane.add(keyCombo, keyCon);

        GridBagConstraints connectCon = new GridBagConstraints();
        connectCon.anchor = GridBagConstraints.LINE_END;
        connectCon.gridx = 3;
        connectCon.gridy = 0;
        connectCon.weightx = 0;
        connectCon.insets = new Insets(15, 20, 5, 5);
        connectButton = new JButton();
        connectButton.setText("Connect");
        connectButton.addActionListener(new ActionListener() {
            @Override
            public void actionPerformed(ActionEvent e) {
                connect();
            }
        });
        frame.getRootPane().setDefaultButton(connectButton);
        pane.add(connectButton, connectCon);

        GridBagConstraints disconnectCon = new GridBagConstraints();
        disconnectCon.anchor = GridBagConstraints.LINE_START;
        disconnectCon.gridx = 4;
        disconnectCon.gridy = 0;
        disconnectCon.weightx = 0;
        disconnectCon.insets = new Insets(15, 15, 5, 5);
        disconnectButton = new JButton();
        disconnectButton.setText("Disconnect");
        disconnectButton.addActionListener(new ActionListener() {
            @Override
            public void actionPerformed(ActionEvent e) {
                disconnect();
            }
        });
        pane.add(disconnectButton, disconnectCon);

        /*
         * OTP checkbox
         */
        GridBagConstraints otpCheckCon = new GridBagConstraints();
        otpCheckCon.anchor = GridBagConstraints.WEST;
        otpCheckCon.gridx = 5;
        otpCheckCon.gridy = 0;
        otpCheckCon.weightx = 0;
        otpCheckCon.insets = new Insets(15, 15, 5, 0);
        otpCheck = new JCheckBox("Authenticate with OTP", false);
        otpCheck.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				if(otpCheck.isEnabled()) {
					if(otpCheck.isSelected()) {
						otpIndivCheck.setEnabled(true);
					} else {
						otpIndivCheck.setEnabled(false);
					}
				}
			}
        });
        pane.add(otpCheck, otpCheckCon);

        /*
         * OTP Individual CheckBox
         */
        GridBagConstraints otpIndivCheckCon = new GridBagConstraints();
        otpIndivCheckCon.anchor = GridBagConstraints.WEST;
        otpIndivCheckCon.gridx = 5;
        otpIndivCheckCon.gridy = 1;
        otpIndivCheckCon.weightx = 0;
        otpIndivCheckCon.insets = new Insets(0, 15, 5, 0);
        otpIndivCheck = new JCheckBox("OTP Individual Account", false);
        otpIndivCheck.setEnabled(false);
        pane.add(otpIndivCheck, otpIndivCheckCon);

        txtProxy = new JTextPane();
        txtProxy.setEditable(false);
        txtProxy.setText("Proxy:");
        GridBagConstraints gbc_txtProxy = new GridBagConstraints();
        gbc_txtProxy.anchor = GridBagConstraints.EAST;
        gbc_txtProxy.insets = new Insets(0, 0, 5, 5);
        gbc_txtProxy.fill = GridBagConstraints.VERTICAL;
        gbc_txtProxy.gridx = 0;
        gbc_txtProxy.gridy = 2;
        frame.getContentPane().add(txtProxy, gbc_txtProxy);

        proxyUrlCombo = new JComboBox();
        proxyUrlCombo.setEditable(true);
        proxyUrlCombo.setToolTipText("proxy url");
        GridBagConstraints gbc_proxyUrlCombo = new GridBagConstraints();
        gbc_proxyUrlCombo.gridwidth = 2;
        gbc_proxyUrlCombo.insets = new Insets(0, 0, 5, 5);
        gbc_proxyUrlCombo.fill = GridBagConstraints.HORIZONTAL;
        gbc_proxyUrlCombo.gridx = 1;
        gbc_proxyUrlCombo.gridy = 2;
        frame.getContentPane().add(proxyUrlCombo, gbc_proxyUrlCombo);

        proxyUsernameCombo = new JComboBox();
        proxyUsernameCombo.setEditable(true);
        proxyUsernameCombo.setToolTipText("proxy username");
        GridBagConstraints gbc_proxyUsernameCombo = new GridBagConstraints();
        gbc_proxyUsernameCombo.insets = new Insets(0, 0, 5, 5);
        gbc_proxyUsernameCombo.fill = GridBagConstraints.HORIZONTAL;
        gbc_proxyUsernameCombo.gridx = 3;
        gbc_proxyUsernameCombo.gridy = 2;
        frame.getContentPane().add(proxyUsernameCombo, gbc_proxyUsernameCombo);

        proxyPasswordTxt = new JTextField();
        proxyPasswordTxt.setToolTipText("proxy password");
        GridBagConstraints gbc_proxyPasswordTxt = new GridBagConstraints();
        gbc_proxyPasswordTxt.insets = new Insets(0, 0, 5, 5);
        gbc_proxyPasswordTxt.fill = GridBagConstraints.HORIZONTAL;
        gbc_proxyPasswordTxt.gridx = 4;
        gbc_proxyPasswordTxt.gridy = 2;
        frame.getContentPane().add(proxyPasswordTxt, gbc_proxyPasswordTxt);
        proxyPasswordTxt.setColumns(10);
        /*
         * Enable Greeks CheckBox
         */
        GridBagConstraints greeksCheckCon = new GridBagConstraints();
        greeksCheckCon.anchor = GridBagConstraints.WEST;
        greeksCheckCon.gridx = 5;
        greeksCheckCon.gridy = 2;
        greeksCheckCon.weightx = 0;
        greeksCheckCon.insets = new Insets(0, 15, 5, 0);
        greeksCheck = new JCheckBox("Enable Greeks", true);
        pane.add(greeksCheck, greeksCheckCon);

        /*
         * Request Button to request new ticker
         */
        GridBagConstraints requestCon = new GridBagConstraints();
        requestCon.anchor = GridBagConstraints.LINE_END;
        requestCon.gridx = 3;
        requestCon.gridy = 1;
        requestCon.insets = new Insets(0, 20, 5, 5);
        requestButton = new JButton();
        requestButton.setText("Request");
        requestButton.addActionListener(new ActionListener() {
            @Override
            public void actionPerformed(ActionEvent e) {
                request();
            }
        });
        pane.add(requestButton, requestCon);

        GridBagConstraints cancelCon = new GridBagConstraints();
        cancelCon.anchor = GridBagConstraints.LINE_START;
        cancelCon.gridx = 4;
        cancelCon.gridy = 1;
        cancelCon.ipadx = 25;
        cancelCon.insets = new Insets(0, 15, 5, 5);
        cancelButton = new JButton();
        cancelButton.setText("Cancel");
        cancelButton.addActionListener(new ActionListener() {
            @Override
            public void actionPerformed(ActionEvent e) {
                cancel();
            }
        });
        pane.add(cancelButton, cancelCon);

        msgTable = new SwingQuoteTable();
        JScrollPane scrollPane = new JScrollPane(msgTable);

        GridBagConstraints tableCon = new GridBagConstraints();
        tableCon.fill = GridBagConstraints.BOTH;
        tableCon.gridx = 0;
        tableCon.gridy = 3;
        tableCon.ipady = 180;
        tableCon.gridwidth = 6;
        tableCon.weightx = 0.1;
        tableCon.weighty = 0.1;
        tableCon.insets = new Insets(0, 0, 5, 0);
        pane.add(scrollPane, tableCon);

        GridBagConstraints statusCon = new GridBagConstraints();
        statusCon.insets = new Insets(0, 0, 5, 5);
        statusCon.fill = GridBagConstraints.HORIZONTAL;
        statusCon.gridx = 0;
        statusCon.gridy = 4;
        statusCon.gridwidth = 5;
        statusLabel = new JLabel();
        pane.add(statusLabel, statusCon);

        GridBagConstraints updatesCon = new GridBagConstraints();
        updatesCon.insets = new Insets(0, 0, 0, 5);
        updatesCon.fill = GridBagConstraints.HORIZONTAL;
        updatesCon.gridx = 0;
        updatesCon.gridy = 5;
        updatesCon.gridwidth = 5;
        updatesLabel = new JLabel();
        pane.add(updatesLabel, updatesCon);

        uriTabPane.setSelectedIndex(prefs.getInt("TAB", 0));
        resetGUIState();

        GridBagConstraints closeCon = new GridBagConstraints();
        closeCon.anchor = GridBagConstraints.LINE_END;
        closeCon.insets = new Insets(0, 0, 10, 0);
        closeCon.gridx = 5;
        closeCon.gridy = 5;
        closeButton = new JButton();
        closeButton.setText("Close");
        closeButton.addActionListener(new ActionListener() {
            @Override
            public void actionPerformed(ActionEvent e) {
                frame.dispose();
            }
        });
        pane.add(closeButton, closeCon);
    }

    private static void addItemToCombo(JComboBox combo, String item) {
        for (int i = 0; i < combo.getItemCount(); ++i) {
            if (combo.getItemAt(i).equals(item)) {
                combo.removeItemAt(i);
                break;
            }
        }
        combo.insertItemAt(item, 0);
        combo.setSelectedIndex(0);
    }

    private void getItemsFromPrefs(JComboBox combo, String key, int numItems) {
        for (int i = 0; i < numItems; ++i) {
            String item = prefs.get(key + "_" + i, "");
            if (!item.isEmpty())
                combo.addItem(prefs.get(key + "_" + i, ""));
        }
    }

    private void putItemsInPrefs(JComboBox combo, String key, int numItems) {
        for (int i = 0; (i < numItems && i < combo.getItemCount()); ++i) {
            prefs.put(key + "_" + i, (String) combo.getItemAt(i));
        }
    }

    private ImageIcon loadIcon(String name) {
        URL iconURL = getClass().getResource(name);
        if (iconURL == null) {
            logger.warning("Error loading icon: " + name);
            return new ImageIcon();
        }
        else
            return new ImageIcon(iconURL);
    }

    private RT_FieldMap loadFieldMap(String uri) {
        try {
            return new RT_FieldMap(uri);
        } catch (SAXException e) {
            logger.log(Level.WARNING, "Error parsing field map", e);
            JOptionPane.showMessageDialog(frame,
                    "Error parsing field map file: " + uri);
            return null;
        } catch (IOException e) {
            logger.log(Level.WARNING, "Error opening field map", e);
            JOptionPane.showMessageDialog(frame,
                    "Error opening field map file: " + uri);
            return null;
        }
    }

    private void createAndShowGUI(String[] args) {
        blankIcon = loadIcon("/resource/blank.gif");
        greenIcon = loadIcon("/resource/green.gif");
        greyIcon  = loadIcon("/resource/grey.gif");

        /*
         * Setting some defaults to avoid Swing making all texts bold by default
         */
        UIManager.put("Label.font", new Font("Tahoma", Font.PLAIN, 11));
        UIManager.put("ComboBox.font", new Font("Tahoma", Font.PLAIN, 11));
        
        /*
         * Create and set up the window.
         */
        frame = new JFrame("SwingQuote");
        frame.setDefaultCloseOperation(WindowConstants.DISPOSE_ON_CLOSE);
        frame.addWindowListener(new WindowAdapter() {
            @Override
            public void windowClosed(WindowEvent we) {
                if (consumer != null)
                    consumer.disconnect();
                putItemsInPrefs(enterpriseUriCombo, "ENT_URI", 3);
                putItemsInPrefs(workstationUriCombo, "WKS_URI", 3);
                putItemsInPrefs(keyCombo, "KEY", 3);
                prefs.putInt("TAB", uriTabPane.getSelectedIndex());
            }
        });
        addComponentsToPane();

        /*
         * Display the window.
         */
        frame.pack();
        frame.setVisible(true);

        try {
            FileHandler fh = new FileHandler("SwingQuote.log");
            fh.setFormatter(new SimpleFormatter());
            logger.addHandler(fh);
            logger.setUseParentHandlers(false);
            Logger.getLogger("com.factset.rt").addHandler(fh);
            Logger.getLogger("com.factset.rt").setUseParentHandlers(false);
        } catch (IOException e) {
            logger.log(Level.WARNING, "Error opening log file", e);
        }

        if (args.length > 0)
            map = loadFieldMap(args[0]);
        else
            map = loadFieldMap("etc/rt_fields.xml");
    }

    public static void main(final String[] args) {
        /*
         * Schedule a job for the event-dispatching thread:
         * creating and showing this application's GUI.
         */
        javax.swing.SwingUtilities.invokeLater(new Runnable() {
            @Override
            public void run() {
                SwingQuote main = new SwingQuote();
                main.createAndShowGUI(args);
            }
        });
    }

    // Helper functions to enable (or disable) proxy items
    private void setProxyItems(boolean is_enable) {
        txtProxy.setEnabled(is_enable);
        proxyUrlCombo.setEnabled(is_enable);
        proxyUsernameCombo.setEnabled(is_enable);
        proxyPasswordTxt.setEnabled(is_enable);
    }
}
